/*
 *	emu.c
 *
 *	Intellivision/EMU
 *
 *	* Mon May 23 00:00:00 JST 2005 Naoyuki Sawa
 *	- 쐬JnB
 */
#include "app.h"

#define AY38914_CLOCK	(NTSC_COLOR_SUBCARRIER/2)
#define AY38914BUFLEN	320

/****************************************************************************
 *	
 ****************************************************************************/

#define KEY_1		 0
#define KEY_2		 1
#define KEY_3		 2
#define KEY_TL		 3
#define KEY_4		 4
#define KEY_5		 5
#define KEY_6		 6
#define KEY_TR		 7
#define KEY_BL		 8
#define KEY_7		 9
#define KEY_8		10
#define KEY_9		11
#define KEY_BR		12
#define KEY_C		13
#define KEY_0		14
#define KEY_E		15
#define KEY_COUNT	16

const unsigned char tenkey_key[KEY_COUNT] = {
	                     INTV_KEY_1    ,INTV_KEY_2,INTV_KEY_3    ,
	INTV_KEY_ACTION_TOP ,INTV_KEY_4    ,INTV_KEY_5,INTV_KEY_6    ,INTV_KEY_ACTION_TOP  ,
	INTV_KEY_ACTION_LEFT,INTV_KEY_7    ,INTV_KEY_8,INTV_KEY_9    ,INTV_KEY_ACTION_RIGHT,
	                     INTV_KEY_CLEAR,INTV_KEY_0,INTV_KEY_ENTER,
};

const char tenkey_pos[KEY_COUNT][2] = {
#define X 41
#define Y  4
	             {X+12,Y+14},{X+24,Y+14},{X+36,Y+14},
	{X+0-3,Y+25},{X+12,Y+25},{X+24,Y+25},{X+36,Y+25},{X+48,Y+25},
	{X+0-3,Y+36},{X+12,Y+36},{X+24,Y+36},{X+36,Y+36},{X+48,Y+36},
	             {X+12,Y+47},{X+24,Y+47},{X+36,Y+47},
#undef X
#undef Y
};

const unsigned char tenkey_dn[KEY_COUNT] = {
	       KEY_4,KEY_5,KEY_6,
	KEY_BL,KEY_7,KEY_8,KEY_9,KEY_BR,
	KEY_TL,KEY_C,KEY_0,KEY_E,KEY_TR,
	       KEY_1,KEY_2,KEY_3,
};

const unsigned char tenkey_up[KEY_COUNT] = {
	       KEY_C,KEY_0,KEY_E,
	KEY_BL,KEY_1,KEY_2,KEY_3,KEY_BR,
	KEY_TL,KEY_4,KEY_5,KEY_6,KEY_TR,
	       KEY_7,KEY_8,KEY_9,
};

const unsigned char tenkey_ri[KEY_COUNT] = {
	      KEY_2,KEY_3,KEY_1 ,
	KEY_4,KEY_5,KEY_6,KEY_TR,KEY_TL,
	KEY_7,KEY_8,KEY_9,KEY_BR,KEY_BL,
	      KEY_0,KEY_E,KEY_C ,
};

const unsigned char tenkey_lf[KEY_COUNT] = {
	       KEY_3 ,KEY_1,KEY_2,
	KEY_TR,KEY_TL,KEY_4,KEY_5,KEY_6,
	KEY_BR,KEY_BL,KEY_7,KEY_8,KEY_9,
	       KEY_E ,KEY_C,KEY_0,
};

int config_key;
int arrow_x;
int arrow_y;

void
do_config()
{
	unsigned char* vbuff_save/*[DISP_X * DISP_Y]*/;
	int old_key;

	vbuff_save = malloc(DISP_X * DISP_Y);
	if(!vbuff_save) {
		DIE();
	}
	memcpy(vbuff_save, vbuff, DISP_X * DISP_Y);

	arrow_x = tenkey_pos[config_key][0];
	arrow_y = tenkey_pos[config_key][1];

	for(;;) {
		schedule();
		memcpy(vbuff, vbuff_save, DISP_X * DISP_Y);
		sprite_draw(0, 0, SPR_CONFIG, 0);

		old_key = config_key;
		if(joy & TRG_UP) config_key = tenkey_up[config_key];
		if(joy & TRG_DN) config_key = tenkey_dn[config_key];
		if(joy & TRG_LF) config_key = tenkey_lf[config_key];
		if(joy & TRG_RI) config_key = tenkey_ri[config_key];
		if(config_key != old_key) {
			sound_play(3, SE_001);
		}

		arrow_x = (arrow_x + tenkey_pos[config_key][0]) / 2;
		arrow_y = (arrow_y + tenkey_pos[config_key][1]) / 2;
		sprite_draw(arrow_x, arrow_y, (now & 4) ? SPR_ARROW0 : SPR_ARROW1,
			(config_key == KEY_TL) || (config_key == KEY_BL) ? DRW_REVX : 0);

		if(joy & (TRG_A|TRG_B|TRG_START|TRG_SELECT)) {
			break;
		}
	}

	free(vbuff_save);

	schedule(); /* gK̂Ă邽 */
}

/****************************************************************************
 *	
 ****************************************************************************/

void
input_update()
{
	int pad = pcePadGetDirect();
	int input = 0;

	if(pad & PAD_UP) {
		if(pad & PAD_LF) {
			input |= INTV_KEY_UP_LEFT;
		} else if(pad & PAD_RI) {
			input |= INTV_KEY_UP_RIGHT;
		} else {
			input |= INTV_KEY_UP;
		}
	} else if(pad & PAD_DN) {
		if(pad & PAD_LF) {
			input |= INTV_KEY_DOWN_LEFT;
		} else if(pad & PAD_RI) {
			input |= INTV_KEY_DOWN_RIGHT;
		} else {
			input |= INTV_KEY_DOWN;
		}
	} else {
		if(pad & PAD_LF) {
			input |= INTV_KEY_LEFT;
		} else if(pad & PAD_RI) {
			input |= INTV_KEY_RIGHT;
		}
	}
	if(pad & PAD_START) {
		input |= INTV_KEY_ENTER;
	}
	if(pad & PAD_A) {
		input |= INTV_KEY_ACTION_TOP;
	}
	if(pad & PAD_B) {
		input |= tenkey_key[config_key];
	}

	ram1[0xfe] = -1;	/* Right */
	ram1[0xff] = ~input;	/* Left  */
}

/****************************************************************************
 *	
 ****************************************************************************/

int
intv_init(const char* fname)
{
	FILE* fp;
	int i;
	int hi;
	int lo;

	/* L[RtBO܂B */
	config_key = KEY_C;

	/* NA܂B */
	memset(ram1, 0, sizeof ram1);
	memset(ram2, 0, sizeof ram2);
	memset(exec, 0, sizeof exec);
	memset(gmem, 0, sizeof gmem);
	memset(cart, 0, sizeof cart);

	//-------------------------------------------------------------------
	fp = fopen("exec.bin", "rb");
	if(!fp) {
		goto L_ERR;
	}
	for(i = 0; i < 0x1000; i++) {
		hi = fgetc(fp);
		lo = fgetc(fp);
		if((hi == EOF) || (lo == EOF)) DIE();
		exec[i] = hi << 8 | lo;
	}
	fclose(fp);
	//-------------------------------------------------------------------
	fp = fopen("grom.bin", "rb");
	if(!fp) {
		goto L_ERR;
	}
	for(i = 0; i < 0x0800; i++) {
		lo = fgetc(fp);
		if(lo == EOF) DIE();
		gmem[i] = lo;
	}
	fclose(fp);
	//-------------------------------------------------------------------
	fp = fopen(fname, "rb");
	if(!fp) {
		goto L_ERR;
	}
	for(i = 0; i < 0x1000; i++) {
		hi = fgetc(fp);
		lo = fgetc(fp);
		if((hi == EOF) || (lo == EOF)) break;
		cart[i] = hi << 8 | lo;
	}
	fclose(fp);
	//-------------------------------------------------------------------

	/* STICZbg܂B */
	stic_reset(ram2, gmem);

	/* PSGZbg܂B */
	ay38910_reset(&ay38910, AY38914_CLOCK, ay38910_in, ay38910_out);

	/* CPUZbg܂B */
	cp1610_reset(0x1000);

	return 0;
L_ERR:
	return -1;
}

/****************************************************************************
 *	
 ****************************************************************************/

int
intv_stream_callback(short* wbuff, int param)
{
	ay38910_mix(&ay38910, wbuff, AY38914BUFLEN);
	return 0;
}

/****************************************************************************
 *	
 ****************************************************************************/

int
intv_run(const char* fname, int sound_on)
{
	int cnt = 0;
	//
	unsigned char* draw;

	if(intv_init(fname) != 0) {
		return -1;
	}

	if(sound_on) {
		stream_play(AY38914BUFLEN, intv_stream_callback, 0, 0);
	}

	cnt = 0;
	for(;;) {
		/* ͍XVB */
		input_update();

		/* t[XLbvB */
		draw = NULL;
		if((cnt += FPS) >= 60) {
			cnt -= 60;
			draw = vbuff;
		}

		//-----------------------------------------------------------
		cp1610_run(INTV_CLOCK/60);
		cp1610_irq(0x1004);
		stic_update(draw);
		//-----------------------------------------------------------

		/* ʍXVB */
		if(draw) {
			stic_reduce(draw, draw);
			schedule();
			if(joy & TRG_START) {
				if(joy & PAD_SELECT) {
					goto L_EXIT;
				}
			}
			if(joy & TRG_SELECT) {
				for(;;) {
					schedule();
					if(joy & PAD_START) {
						goto L_EXIT;
					}
					if(!(joy & PAD_SELECT)) {
						if(sound_on) { /* TEhꎞ~BL[ʉ炷̂ŁAOff͕s */
							stream_stop();
						}
						do_config();
						if(sound_on) { /* TEhA */
							stream_play(AY38914BUFLEN, intv_stream_callback, 0, 0);
						}
						break;
					}
				}
			}
		}
	}

L_EXIT:
	if(sound_on) {
		stream_stop();
	}

	return 0;
}

/****************************************************************************
 *	
 ****************************************************************************/

/* AY-3-8914̃AhXItZbgAAY-3-8910̃|[gԍւ̕ϊ\ */
const unsigned char psg_map[16] = {
	 0,	/* Channel A Tone Period (Fine Tune) */
	 2,	/* Channel B Tone Period (Fine Tune) */
	 4,	/* Channel C Tone Period (Fine Tune) */
	11,	/* Envelope Period (Fine Tune) */
	 1,	/* Channel A Tone Period (Coarse Tune) */
	 3,	/* Channel B Tone Period (Coarse Tune) */
	 5,	/* Channel C Tone Period (Coarse Tune) */
	12,	/* Envelope Period (Coarse Tune) */
	 7,	/* #Enable */
	 6,	/* Noise Period */
	13,	/* Envelope Shape Cycle */
	 8,	/* Channel A Amplitude */
	 9,	/* Channel B Amplitude */
	10,	/* Channel C Amplitude */
	14,	/* I/O Port A Data Store */
	15,	/* I/O Port B Data Store */
};

int
ay38914_read(int addr)
{
	int data;
	ASSERT((addr >= 0) && (addr <= 0x0f));

	addr = psg_map[addr];
ENTER_CS;
	ay38910_address(&ay38910, addr);
	data= ay38910_read(&ay38910);
LEAVE_CS;
	return data;
}

void
ay38914_write(int addr, int data)
{
	ASSERT((addr >= 0) && (addr <= 0x0f));
	ASSERT((data >= 0) && (addr <= 0xff));

	addr = psg_map[addr];
ENTER_CS;
	ay38910_address(&ay38910, addr);
	ay38910_write(&ay38910, data);
LEAVE_CS;
}

unsigned char
ay38910_in(AY38910* ay38910, int port)
{
	ASSERT((port >= 0) && (port <= 1));

	return ram1[0xfe + port];
}

void
ay38910_out(AY38910* ay38910, int port, unsigned char data)
{
	/** no job **/
}

/****************************************************************************
 *	
 ****************************************************************************/

